-- Cuboid helper functions
rawset(_G, "cuboidhelper", {
	create = function(frames, xsize, ysize, zsize, xpivot, ypivot, zpivot, scale, mirror, rollmode, x, y, z, xyangle, zangle)
		local cuboid = {
			widefaces = {},
			longfaces = {},
			
			frames = frames,
			
			xsize = xsize,
			ysize = ysize,
			zsize = zsize,
			xpivot = xpivot,
			ypivot = ypivot,
			zpivot = zpivot,
			
			scale = scale,
			mirror = mirror,
			rollmode = rollmode,
			
			x = x,
			y = y,
			z = z,
			xyangle = xyangle,
			zangle = zangle
		}
		
		for i = 1, 4 do
			local face = P_SpawnMobj(0, 0, 0, MT_MCFACE)
			face.origframe = frames[i]
			face.renderflags = RF_SLOPESPLAT | RF_NOSPLATBILLBOARD | RF_FLOORSPRITE
			face.spriteyoffset = -4*FRACUNIT
			P_CreateFloorSpriteSlope(face)
			cuboid.widefaces[i] = face
		end
		
		for i = 1, 2 do
			local face = P_SpawnMobj(0, 0, 0, MT_MCFACE)
			face.origframe = frames[i + 4]
			face.renderflags = RF_PAPERSPRITE
			cuboid.longfaces[i] = face
		end
		
		return cuboid
	end,
	
	update = function(cuboid, realOnly)
		cuboid.real = {
			x = cuboid.x,
			y = cuboid.y,
			z = cuboid.z,
			xyangle = cuboid.xyangle + ANGLE_180,
			zangle = -cuboid.zangle,
			xsize = FixedMul(cuboid.scale, cuboid.xsize),
			ysize = FixedMul(cuboid.scale, cuboid.ysize),
			zsize = FixedMul(cuboid.scale, cuboid.zsize),
			xpivot = cuboid.xpivot,
			ypivot = -cuboid.ypivot
		}
		
		local sign = cuboid.mirror and -1 or 1
		
		if cuboid.rollmode then
			cuboid.real.xyangle = $ - ANGLE_90 * sign
			cuboid.real.zangle = $ * sign
			cuboid.real.xsize, cuboid.real.ysize = $2, $1
			cuboid.real.xpivot, cuboid.real.ypivot = $2 * sign, -$1 * sign
		end
		
		for i = 0, 3 do
			if cuboid.real.zangle > ANGLE_90 * i and cuboid.real.zangle < ANGLE_90 * i + 256 then
				cuboid.real.zangle = ANGLE_90 * i
				break
			end
		end
		
		if cuboid.real.xpivot != 0 then
			cuboid.real.x = $ + FixedMul(sin(cuboid.real.xyangle), -cuboid.real.xpivot)
			cuboid.real.y = $ + FixedMul(cos(cuboid.real.xyangle), cuboid.real.xpivot)
		end
		if cuboid.real.ypivot != 0 then
			cuboid.real.x = $ + FixedMul(cos(cuboid.real.xyangle), cuboid.real.ypivot)
			cuboid.real.y = $ + FixedMul(sin(cuboid.real.xyangle), cuboid.real.ypivot)
		end
		if cuboid.zpivot != 0 then
			cuboid.real.x = $ + FixedMul(sin(cuboid.real.zangle), FixedMul(cos(cuboid.real.xyangle), cuboid.zpivot))
			cuboid.real.y = $ + FixedMul(sin(cuboid.real.zangle), FixedMul(sin(cuboid.real.xyangle), cuboid.zpivot))
			cuboid.real.z = $ + FixedMul(cos(cuboid.real.zangle), -cuboid.zpivot)
		end
		
		if realOnly then return end
		
		for _, faces in ipairs({cuboid.widefaces, cuboid.longfaces}) do
			for _, face in ipairs(faces) do face.frame = face.origframe end
		end
		
		if cuboid.rollmode then
			cuboid.widefaces[1].frame, cuboid.widefaces[3].frame, cuboid.longfaces[1].frame, cuboid.longfaces[2].frame =  $4, $3, $1, $2
		end
		
		for i, face in ipairs(cuboid.widefaces) do
			face.frame = $ & ~FF_PAPERSPRITE | FF_FLOORSPRITE
			if i == 2 or i == 3 then face.frame = $ ^^ FF_VERTICALFLIP end
			if i == 3 then face.frame = $ ^^ FF_HORIZONTALFLIP end
		end
		for _, face in ipairs(cuboid.longfaces) do
			face.frame = $ & ~FF_FLOORSPRITE | FF_PAPERSPRITE
		end
		
		local zanglefixed = AngleFixed(cuboid.real.zangle)
		for i, face in ipairs(cuboid.widefaces) do
			local widezangle = cuboid.real.zangle + ANGLE_90 * i
			local widezanglefixed = AngleFixed(widezangle)
			
			if widezangle == ANGLE_90 or widezangle == ANGLE_270 then
				face.floorspriteslope.xydirection = cuboid.real.xyangle
				face.floorspriteslope.zangle = ANGLE_90 - 1
				if widezangle == ANGLE_270 then face.frame = $ ^^ FF_VERTICALFLIP end
			elseif widezanglefixed > 270*FRACUNIT then
				face.floorspriteslope.xydirection = cuboid.real.xyangle + ANGLE_180
				face.floorspriteslope.zangle = -widezangle
			elseif widezanglefixed > 180*FRACUNIT then
				face.floorspriteslope.xydirection = cuboid.real.xyangle
				face.floorspriteslope.zangle = abs(ANGLE_180 - widezangle)
				face.frame = $ ^^ FF_VERTICALFLIP
			elseif widezanglefixed > 90*FRACUNIT then
				face.floorspriteslope.xydirection = cuboid.real.xyangle + ANGLE_180
				face.floorspriteslope.zangle = abs(widezangle - ANGLE_180)
				face.frame = $ ^^ FF_VERTICALFLIP
			else
				face.floorspriteslope.xydirection = cuboid.real.xyangle
				face.floorspriteslope.zangle = widezangle
			end
			
			face.angle = cuboid.real.xyangle
			face.spritexscale = cuboid.scale
			face.spriteyscale = FixedMul(cos(face.floorspriteslope.zangle), cuboid.scale + (face.floorspriteslope.zangle == ANGLE_90 - 1 and 1024 or 0))
			
			if cuboid.mirror then face.frame = $ ^^ FF_HORIZONTALFLIP end
			if cuboid.rollmode and i % 2 == 0 then
				face.angle = $ + ANGLE_90 * sign
				face.spritexscale, face.spriteyscale = face.spriteyscale, face.spritexscale
			end
			
			local wideoffset = (i % 2 == 1 and cuboid.real.ysize or cuboid.real.zsize) / 2
			P_MoveOrigin(face,
				cuboid.real.x + FixedMul(cos(cuboid.real.xyangle), FixedMul(sin(widezangle), -wideoffset)),
				cuboid.real.y + FixedMul(sin(cuboid.real.xyangle), FixedMul(sin(widezangle), -wideoffset)),
				cuboid.real.z + FixedMul(cos(widezangle),  wideoffset) + FixedMul(4*FRACUNIT, cuboid.scale)
			)
			
			face.floorspriteslope.o = {face.x, face.y, face.z}
		end
		
		local function adjustpos(center) return FixedMul(cuboid.scale, max(0, min(FRACUNIT, -abs((zanglefixed - center*FRACUNIT) / 90) + (3*FRACUNIT)/2))) end
		local longxyoffset = cuboid.real.xsize / 2
		local longxyangle = cuboid.real.xyangle + ANGLE_90
		local longxyadjust = adjustpos(cuboid.mirror and 225 or 135) * sign
		local longzadjust = adjustpos(cuboid.mirror and 135 or 225)
		for i, face in ipairs(cuboid.longfaces) do
			face.angle = cuboid.real.xyangle
			if cuboid.mirror then face.frame = $ ^^ FF_HORIZONTALFLIP end
			face.spritexscale = cuboid.scale
			face.spriteyscale = cuboid.scale
			face.spriteroll = cuboid.real.zangle * sign
			
			local sign = (i * 2 - 3) * sign
			P_MoveOrigin(face,
				cuboid.real.x + FixedMul(longxyoffset, cos(longxyangle)) * sign + FixedMul(longxyadjust, cos(cuboid.real.xyangle)),
				cuboid.real.y + FixedMul(longxyoffset, sin(longxyangle)) * sign + FixedMul(longxyadjust, sin(cuboid.real.xyangle)),
				cuboid.real.z + longzadjust
			)
		end
	end,
	
	destroy = function(cuboid)
		for _, faces in ipairs({cuboid.widefaces, cuboid.longfaces}) do
			for _, face in ipairs(faces) do
				if face.valid then P_RemoveMobj(face) end
			end
		end
		cuboid = nil
	end,
	
	modifywide = function(cuboid, modifier) for _, face in ipairs(cuboid.widefaces) do modifier(face) end end,
	modifylong = function(cuboid, modifier) for _, face in ipairs(cuboid.longfaces) do modifier(face) end end,
	
	validate = function(cuboid)
		if #cuboid.widefaces != 4 or #cuboid.longfaces != 2 then return false end
		for _, faces in ipairs({cuboid.widefaces, cuboid.longfaces}) do
			for _, face in ipairs(faces) do
				if not face.valid then return false end
			end
		end
		return true
	end
})