-- Custom Minecraft-style HUD
stevehelper.customhud = function(screen, player)
	if stevehelper.validinfo(player) then
		-- Cache normal and blinking heart graphics
		local fullheart = {
			screen.cachePatch("MCHTFULA"),
			screen.cachePatch("MCHTFULB")
		}
		local halfheart = {
			screen.cachePatch("MCHTHLFA"),
			screen.cachePatch("MCHTHLFB")
		}
		local noneheart = {
			screen.cachePatch("MCHTNONA"),
			screen.cachePatch("MCHTNONB"),
		}
		
		local steve = player.steve
		
		-- Render health bar
		local health = steve.persist.health
		local oldhealth = steve.play.oldhealth
		local blinktimer = steve.play.blinktimer
		local regentimer = steve.play.regentimer
		local blinkmode = (blinktimer > 0 and (blinktimer / 4) % 2 == 0) and 2 or 1
		for heart = 0, 9 do
			local heart2 = heart * 2
			local x = 69 + heart * 8
			local y = 168 - ((regentimer != 0 and heart + 1 == 10 - regentimer / 2) and 2 or 0)
			screen.draw(x, y, noneheart[blinkmode], V_SNAPTOBOTTOM)
			if blinkmode == 2 then
				if heart2 < oldhealth then
					if heart2 == oldhealth - 1 then
						screen.draw(x, y, halfheart[2], V_SNAPTOBOTTOM)
					else
						screen.draw(x, y, fullheart[2], V_SNAPTOBOTTOM)
					end
				end
			end
			if heart2 < health then
				if heart2 == health - 1 then
					screen.draw(x, y, halfheart[1], V_SNAPTOBOTTOM)
				else
					screen.draw(x, y, fullheart[1], V_SNAPTOBOTTOM)
				end
			end
		end
		
		if steve.play.submerged then
			-- Cache normal and popping bubble graphics
			local normalbubble = screen.cachePatch("MCBUFULL")
			local poppedbubble = screen.cachePatch("MCBUPOPD")
			
			-- Render breath bubbles
			local breath = (steve.play.breath + (steve.play.breath % 2)) / 2
			local poptimer = steve.play.poptimer
			for bubble = 1, 10 do
				local x = 170 + (10 - bubble) * 8
				if bubble <= breath then
					screen.draw(x, 168, normalbubble, V_SNAPTOBOTTOM)
				elseif steve.play.breath % 2 == 0 and bubble == breath + 1 and poptimer > 0 then
					screen.draw(x, 168, poppedbubble, V_SNAPTOBOTTOM)
				end
			end
		end
		
		-- Render hotbar
		local hotbarslot = steve.persist.hotbarslot
		screen.draw(69, 178, screen.cachePatch("MCHBFRNT"), V_SNAPTOBOTTOM)
		screen.draw(72, 181, screen.cachePatch("MCHBBACK"), V_SNAPTOBOTTOM | V_30TRANS)
		screen.draw(68 + hotbarslot * 20, 177, screen.cachePatch("MCHBSLCT"), V_SNAPTOBOTTOM)
		
		local hotbar = steve.persist.hotbar
		for slot = 0, 8 do
			if hotbar[slot].item == MT_NULL or hotbar[slot].amount <= 0 then continue end
			
			-- Render items
			local itemstate = states[mobjinfo[hotbar[slot].item].spawnstate]
			local patch = screen.getSpritePatch(itemstate.sprite, itemstate.frame, 3)
			if patch == nil or not patch.valid then patch = screen.getSpritePatch(SPR_UNKN) end
			local scale = FixedDiv(16*FRACUNIT, max(patch.width, patch.height)*FRACUNIT)
			local xoffset = (16*FRACUNIT - patch.width * scale) / 2
			local yoffset = (16*FRACUNIT - patch.height * scale) / 2
			screen.drawScaled(
				(72 + 20 * slot)*FRACUNIT + patch.leftoffset * scale + xoffset,
				181*FRACUNIT + patch.topoffset * scale + yoffset,
				scale, patch, V_SNAPTOBOTTOM
			)
			
			-- Render quantity
			local numbers = screen.cachePatch("MCHBNUMS")
			local digit1 = hotbar[slot].amount % 10
			local digit2 = ((hotbar[slot].amount - digit1) / 10) % 10
			if digit1 > 1 or digit2 > 0 then
				for i, digit in ipairs({digit1, digit2}) do
					if i > 1 and digit == 0 then continue end
					screen.drawCropped(
						(83 + 6 * (1 - i) + 20 * slot)*FRACUNIT, 190*FRACUNIT, FRACUNIT, FRACUNIT, numbers,
						V_SNAPTOBOTTOM, nil, 6*FRACUNIT * digit, 0, 6*FRACUNIT, 8*FRACUNIT
					)
				end
			end
		end
	end
end
hud.add(stevehelper.customhud, "game")