-- Steve helper functions
rawset(_G, "stevehelper", {
	create = function(player, scale)
		if player.steve == nil then player.steve = {} end
		local steve = player.steve
		
		if steve.model == nil then
			local argslist = {
				{{A, B, C, D, E, F}, 32*FRACUNIT, 32*FRACUNIT, 32*FRACUNIT, 0, 0, 0, 0, false, false, 0, 0, 0, 0, 0},
				{{G, H, I, J, K, L}, 32*FRACUNIT, 16*FRACUNIT, 48*FRACUNIT, 0, 0, 0, 0, false, false, 0, 0, 0, 0, 0},
				{{M, N, O, P, Q, R}, 16*FRACUNIT, 16*FRACUNIT, 48*FRACUNIT, 0, 0, 0, 0, true,  false, 0, 0, 0, 0, 0},
				{{M, N, O, P, Q, R}, 16*FRACUNIT, 16*FRACUNIT, 48*FRACUNIT, 0, 0, 0, 0, false, false, 0, 0, 0, 0, 0},
				{{S, T, U, V, W, X}, 16*FRACUNIT, 16*FRACUNIT, 48*FRACUNIT, 0, 0, 0, 0, true,  false, 0, 0, 0, 0, 0},
				{{S, T, U, V, W, X}, 16*FRACUNIT, 16*FRACUNIT, 48*FRACUNIT, 0, 0, 0, 0, false, false, 0, 0, 0, 0, 0},
			}
			
			steve.model = {
				main = {},
				mask = {},
				argslist = argslist,
				
				x = player.mo.x,
				y = player.mo.y,
				z = player.mo.z,
				xyangle = player.mo.angle,
				zangle = player.aiming,
				
				bodyxyangle = player.mo.angle,
				bodyzangle = 0,
				scale = scale
			}
			
			for i, args in ipairs(steve.model.argslist) do
				steve.model.main[i] = cuboidhelper.create(unpack(args))
				local modifier = function(face) face.dontdrawforviewmobj = player.mo end
				cuboidhelper.modifywide(steve.model.main[i], modifier)
				cuboidhelper.modifylong(steve.model.main[i], modifier)
			end
		end
		
		if steve.play == nil then
			steve.play = {
				moving = false,
				sneaking = false,
				sprinting = false,
				bhopping = false,
				swimming = false,
				sitting = false,
				hanging = false,
				tossing = false,
				dying = false,
				
				underwater = false,
				goopwater = false,
				submerged = false,
				antigravity = false,
				
				hurttimer = 0,
				redtimer = 0,
				blinktimer = 0,
				regentimer = 0,
				soundtimer = 0,
				
				breath = 20,
				breathoffset = 0,
				poptimer = 0,
				
				usepos = 0,
				usemax = 15,
				usetimer = 0,
				
				punchaction = false,
				punchactive = false,
				placeaction = false,
				placeactive = false,
				
				oldhealth = 20,
				currentfov = 0,
				
				swing = {
					progress = 0,
					progressadd = ANG20,
					extent = 0,
					realextent = 0,
					extentadd = FixedAngle(4*FRACUNIT),
					minextent = ANG20,
					maxextent = ANG60
				},
				bob = {
					progress = 0,
					progressadd = FixedAngle(12*FRACUNIT),
					extent = 0,
					realextent = 0,
					extentadd = FRACUNIT/7,
					minextent = 0,
					maxextent = (FRACUNIT*9)/4
				},
				
				last = {}
			}
		end
		
		if steve.persist == nil then
			steve.persist = {
				health = 20,
				hotbar = {},
				hotbarbackup = {},
				hotbarslot = 0,
				spawntime = 0
			}
			
			local hotbar = steve.persist.hotbar
			local hotbarbackup = steve.persist.hotbarbackup
			for slot = 0, 8 do
				hotbar[slot] = {item = MT_NULL, amount = 0}
				hotbarbackup[slot] = {item = MT_NULL, amount = 0}
			end
		end
	end,
	
	update = function(player)
		local steve = player.steve
		local mainmodel, maskmodel = steve.model.main, steve.model.mask
		local realscale = FixedMul(steve.model.scale, player.mo.scale)
		
		for i, cuboid in ipairs(mainmodel) do
			local modifier = function(face)
				face.color = player.skincolor
				if player.mo.flags2 & MF2_DONTDRAW and steve.play.redtimer == 0 then
					face.flags2 = $ | MF2_DONTDRAW
				else
					face.flags2 = $ & ~MF2_DONTDRAW
				end
			end
			cuboidhelper.modifywide(cuboid, modifier)
			cuboidhelper.modifylong(cuboid, modifier)
			
			cuboid.xpivot, cuboid.ypivot, cuboid.zpivot = 0, 0, 0
			cuboid.scale = realscale + (i != 2 and realscale / 4096 or 0)
			cuboid.rollmode = false
			
			cuboid.x, cuboid.y, cuboid.z = steve.model.x, steve.model.y, steve.model.z
			cuboid.xyangle, cuboid.zangle = steve.model.bodyxyangle, 0
		end
		
		mainmodel[1].zpivot = realscale * -16
		mainmodel[1].z = $ + realscale * 96
		
		mainmodel[2].z = $ + realscale * 72
		
		for i = 3, 4 do
			local sign = i * 2 - 7
			mainmodel[i].xpivot = realscale * (24 * sign)
			mainmodel[i].zpivot = realscale * 16
			mainmodel[i].z = $ + realscale * 88
		end
		
		for i = 5, 6 do
			local sign = i * 2 - 11
			mainmodel[i].xpivot = realscale * (8 * sign)
			mainmodel[i].zpivot = realscale * 24
			mainmodel[i].z = $ + realscale * 48
		end
		
		if not steve.play.dying then
			local xybobstate = FixedMul(sin(steve.play.bob.progress), FixedMul(player.mo.scale, steve.play.bob.extent))
			local zbobstate = FixedMul(-abs(cos(steve.play.bob.progress)), FixedMul(player.mo.scale, steve.play.bob.extent))
			for i, cuboid in ipairs(mainmodel) do
				cuboid.x = $ + FixedMul(cos(steve.model.xyangle + ANGLE_90), xybobstate)
				cuboid.y = $ + FixedMul(sin(steve.model.xyangle + ANGLE_90), xybobstate)
				cuboid.z = $ + zbobstate - (steve.play.sitting and realscale * 36 or 0)
			end
			
			mainmodel[1].xyangle = steve.model.xyangle
			mainmodel[1].zangle = steve.model.zangle * (steve.play.antigravity and -1 or 1)
			
			local armswingstate = FixedMul(sin(steve.play.swing.progress), steve.play.swing.extent)
			for i = 3, 4 do
				local sign = i * 2 - 7
				if steve.play.sitting then
					mainmodel[i].zangle = ANG30
				elseif steve.play.hanging then
					mainmodel[i].zangle = ANGLE_180
				else
					mainmodel[i].zangle = armswingstate * -sign
				end
			end
			
			local legswingstate = (armswingstate / 4) * 5
			for i = 5, 6 do
				local sign = i * 2 - 11
				if steve.play.sitting or steve.play.hanging then
					mainmodel[i].xyangle = $ + ANG15 * -sign
					mainmodel[i].zangle = FixedAngle(80*FRACUNIT)
				else
					mainmodel[i].zangle = legswingstate * sign
				end
			end
			
			if steve.play.sneaking then
				mainmodel[1].z = $ + realscale * -10
				
				mainmodel[2].ypivot = realscale * -8
				mainmodel[2].z = $ + realscale * -8
				mainmodel[2].zangle = ANGLE_337h
				
				for i = 3, 4 do
					mainmodel[i].z = $ + realscale * -8
					if not steve.play.sitting and not steve.play.hanging then
						mainmodel[i].zangle = $ - ANG15
					end
				end
				
				for i = 5, 6 do
					mainmodel[i].ypivot = $ + realscale * -12
				end
			end
			
			if steve.play.usepos > 0 and steve.play.usepos < steve.play.usemax then
				local arm = mainmodel[4]
				
				local initxyangle, initzangle = arm.xyangle, arm.zangle
				arm.xyangle = steve.model.bodyxyangle + (steve.play.antigravity and ANGLE_180 or 0)
				arm.zangle = 0
				cuboidhelper.update(arm, true)
				
				arm.xpivot = 0
				arm.x = arm.real.x
				arm.y = arm.real.y
				arm.xyangle = $ + (steve.play.antigravity and ANGLE_180 or 0)
				
				local progress = FixedDiv(steve.play.usepos*FRACUNIT, steve.play.usemax*FRACUNIT)
				if progress < FRACUNIT/5 then
					local segprogress = progress * 5
					arm.xyangle = $ + ease.linear(segprogress, -ANG60, 0)
					arm.zangle = ease.outsine(segprogress, initzangle, ANGLE_90)
				elseif progress < (FRACUNIT*2)/5 then
					local segprogress = (progress - FRACUNIT/5) * 5
					arm.xyangle = $ + ease.outsine(segprogress, 0, ANG60)
					arm.zangle = ease.insine(segprogress, ANGLE_90, ANGLE_45)
				else
					local segprogress = FixedDiv(progress - (FRACUNIT*2)/5, (FRACUNIT*3)/5)
					arm.xyangle = $ + ease.outsine(segprogress, ANG60, 0)
					arm.zangle = ease.outsine(segprogress, ANGLE_45, initzangle)
				end
			end
		else
			for _, cuboid in ipairs(mainmodel) do cuboid.rollmode = true end
			steve.play.redtimer = -1
			steve.model.bodyzangle = min(ANGLE_90, $ + max(ANG1, (ANGLE_90 - $) / 7))
			if steve.play.antigravity then
				steve.model.z = player.mo.z + realscale * 120
			end
		end
		
		local realbodyzangle = steve.model.bodyzangle + (steve.play.antigravity and ANGLE_180 or 0)
		local modelbackup = {}
		if realbodyzangle != 0 then
			for i, cuboid in ipairs(mainmodel) do
				local x, y, z = cuboid.x, cuboid.y, cuboid.z
				local xpivot, ypivot, zpivot = cuboid.xpivot, cuboid.ypivot, cuboid.zpivot
				local xyangle, zangle = cuboid.xyangle, cuboid.zangle
				
				cuboid.xpivot, cuboid.ypivot, cuboid.zpivot = 0, 0, 0
				cuboid.xyangle = steve.model.bodyxyangle + ((steve.play.antigravity and not cuboid.rollmode) and ANGLE_180 or 0)
				cuboidhelper.update(cuboid, true)
				
				cuboid.z = steve.model.z + ((not steve.play.dying) and realscale * 64 or 0)
				cuboid.zpivot = cuboid.z - cuboid.real.z
				cuboid.zangle = realbodyzangle
				cuboidhelper.update(cuboid, true)
				
				cuboid.x, cuboid.y, cuboid.z = cuboid.real.x, cuboid.real.y, cuboid.real.z
				if cuboid.rollmode then
					cuboid.z = $ + FixedMul(sin(cuboid.zangle), xpivot)
				else
					cuboid.z = $ + FixedMul(sin(cuboid.zangle), ypivot)
				end
				
				cuboid.xpivot, cuboid.ypivot, cuboid.zpivot = xpivot, ypivot, zpivot
				if cuboid.rollmode then
					cuboid.xpivot = FixedMul(cos(cuboid.zangle), $)
				else
					cuboid.ypivot = FixedMul(cos(cuboid.zangle), $)
				end
				
				cuboid.xyangle = $ + (steve.model.bodyxyangle - xyangle) * ((i == 1 and steve.play.antigravity) and -1 or 1)
				cuboid.zangle = $ + zangle
				cuboidhelper.update(cuboid, false)
				
				modelbackup[i] = {x, y, z, xpivot, ypivot, xyangle, zangle}
			end
		else
			for _, cuboid in ipairs(mainmodel) do cuboidhelper.update(cuboid, false) end
		end
		
		if steve.play.redtimer > 0 or steve.play.redtimer < 0 then
			if #maskmodel == 0 then
				for i, args in ipairs(steve.model.argslist) do
					maskmodel[i] = cuboidhelper.create(unpack(args))
					local modifier = function(face)
						face.origframe = $ | FF_TRANS70
						face.color = SKINCOLOR_RED
						face.colorized = true
						face.dontdrawforviewmobj = player.mo
					end
					cuboidhelper.modifywide(maskmodel[i], modifier)
					cuboidhelper.modifylong(maskmodel[i], modifier)
				end
			else
				for _, cuboid in ipairs(maskmodel) do
					local modifier = function(face) face.flags2 = $ & ~MF2_DONTDRAW end
					cuboidhelper.modifywide(cuboid, modifier)
					cuboidhelper.modifylong(cuboid, modifier)
				end
			end
			for i, cuboid in ipairs(maskmodel) do
				for key, value in pairs(mainmodel[i]) do
					if key == "longfaces" or key == "widefaces" or key == "frames" then continue end
					cuboid[key] = value
				end
				cuboidhelper.update(cuboid, false)
			end
		elseif #maskmodel != 0 then
			for _, cuboid in ipairs(maskmodel) do
				local modifier = function(face) face.flags2 = $ | MF2_DONTDRAW end
				cuboidhelper.modifywide(cuboid, modifier)
				cuboidhelper.modifylong(cuboid, modifier)
			end
		end
		
		if realbodyzangle != 0 then
			for i, cuboid in ipairs(mainmodel) do
				cuboid.x, cuboid.y, cuboid.z, cuboid.xpivot, cuboid.ypivot, cuboid.xyangle, cuboid.zangle = unpack(modelbackup[i])
			end
		end
	end,
	
	/*
		Restore the player to a non-Steveified state
		
		Modes:
		0 - Destroy model
		1 - The above + wipe non-persistent player information
		2 - The above + reset health
		3 - The above + wipe persistent player information
		4 - The above + revert player to default state
	*/
	destroy = function(player, mode)
		local steve = player.steve
		if steve == nil then return end
		
		-- Remove Steve model
		if steve.model != nil then
			for _, model in ipairs({steve.model.main, steve.model.mask}) do
				for _, cuboid in ipairs(model) do cuboidhelper.destroy(cuboid) end
			end
			steve.model = nil
		end
		if mode == 0 then return end
		
		-- Remove non-persistent player information
		steve.play = nil
		if mode == 1 then return end
		
		-- Reset health
		if steve.persist != nil then steve.persist.health = 20 end
		if mode == 2 then return end
		
		-- Remove persistent player information
		steve.persist = nil
		if mode == 3 then return end
		
		-- Remove Steve object
		steve = nil
		
		-- Re-enable vanilla HUD
		if #consoleplayer == #player then
			hud.enable("score")
			hud.enable("time")
			hud.enable("rings")
			hud.enable("lives")
		end
		
		-- Reset speed values
		player.normalspeed = skins[player.mo.skin].normalspeed
		player.acceleration = skins[player.mo.skin].acceleration
	end,
	
	-- Ensure that a player is Steve and is treated as such
	validinfo = function(player)
		return player.mo != nil and player.mo.skin == "mcsteve" and player.steve != nil and player.steve.play != nil and player.steve.persist != nil
	end,
	
	-- Ensure that a player has a valid Steve model
	validmodel = function(player)
		if player.steve == nil or player.steve.model == nil then return false end
		for _, model in ipairs({player.steve.model.main, player.steve.model.mask}) do
			for _, cuboid in ipairs(model) do
				if not cuboidhelper.validate(cuboid) then return false end
			end
		end
		return true
	end,
	
	-- Get movement speed relative to maximum speed
	currentspeed = function(player)
		local currentspeed = FixedDiv(FixedHypot(player.rmomx, player.rmomy), player.mo.scale)
		if player.powers[pw_sneakers] or player.powers[pw_super] then
			currentspeed = ($ * 3) / 5
		end
		if P_IsObjectOnGround(player.mo) then
			currentspeed = FixedDiv($, player.mo.friction)
		end
		return currentspeed
	end,
	
	-- Return maximum desired speed based on current state
	maxspeed = function(player)
		local speed = skins[player.mo.skin].normalspeed
		if player.steve.play.sneaking   then speed = $ / 2 end
		if player.steve.play.sprinting  then speed = ($ * 4) / 3 end
		if player.steve.play.bhopping   then speed = ($ * 5) / 4 end
		if player.steve.play.underwater then speed = ($ * 3) / 2 end
		return speed
	end,
	
	-- Deal a specific amount of damage to the player
	damage = function(player, source, amount)
		local steve = player.steve
		
		-- Take health away from player
		steve.play.oldhealth = steve.persist.health
		steve.persist.health = $ - amount
		
		-- Set timers for invincibility, red tint, and health bar blink
		steve.play.hurttimer = 19
		steve.play.redtimer = 10
		steve.play.blinktimer = 15
		
		-- Throw the player back if the damage is caused by another object, and launch upwards if on the ground
		if source != nil and source.valid then
			local throwangle = R_PointToAngle2(player.mo.x, player.mo.y, source.x, source.y) + ANGLE_180
			P_InstaThrust(player.mo, throwangle, player.mo.scale * 7)
			if P_IsObjectOnGround(player.mo) then
				player.mo.z = $ + (steve.play.antigravity and -1 or 1)
				P_SetObjectMomZ(player.mo, 10*FRACUNIT)
			elseif player.mo.momz < 0 then
				player.mo.momz = 0
			end
		end
		
		if steve.persist.health <= 0 then
			-- Kill player if there is no health left
			P_DamageMobj(player.mo, nil, nil, 0, DMG_INSTAKILL)
		else
			-- Play "oof" sound if player is not dead yet
			-- Otherwise the sound is played by the death handler
			S_StartSound(player.mo, sfx_mcoof1 + P_RandomKey(3))
		end
	end,
	
	-- Give the player an item
	giveitem = function(player, type)
		local hotbar = player.steve.persist.hotbar
		for pass = 1, 2 do
			for slot = 0, 8 do
				-- Try to add item to an existing stack during the first pass
				-- Otherwise, try to add item to an empty slot during the second pass
				if (pass == 1 and (hotbar[slot].amount <= 0 or hotbar[slot].item == MT_NULL))
				or (pass == 2 and hotbar[slot].amount > 0 and hotbar[slot].item != MT_NULL) then
					continue
				end
				-- Check if the item can be added, and add it if so
				if hotbar[slot].amount <= 0 or hotbar[slot].item == MT_NULL
				or (hotbar[slot].amount < 64 and hotbar[slot].item == type) then
					hotbar[slot].amount = hotbar[slot].item == MT_NULL and 1 or max(1, $ + 1)
					hotbar[slot].item = type
					if player.steve.play.soundtimer == 0 then
						S_StartSoundAtVolume(player.mo, sfx_mcpop1 + P_RandomKey(3), 96)
						player.steve.play.soundtimer = 1
					end
					return true
				end
			end
		end
		return false
	end,
	
	-- Spawn an item object travelling at a certain angle and velocity
	dropitem = function(source, type, xyangle, zangle, momentum, antigravity)
		local dropheight = 0
		if source.type == MT_PLAYER then
			local steve = source.player.steve
			if not antigravity then
				dropheight = source.z + FixedMul(steve.model.scale, source.scale) * 96
			else
				dropheight = source.z + source.height - FixedMul(steve.model.scale, source.scale) * 96 - FixedMul(source.scale, mobjinfo[MT_MCITEM].height)
			end
			momentum = FixedMul($, source.scale)
		else
			dropheight = source.z + source.height / 2
		end
		
		local item = P_SpawnMobj(source.x, source.y, dropheight, MT_MCITEM)
		local spritescale = FixedDiv(24*FRACUNIT, max(mobjinfo[type].radius * 2, mobjinfo[type].height))
		
		item.momx = FixedMul(momentum, FixedMul(cos(xyangle), cos(zangle)))
		item.momy = FixedMul(momentum, FixedMul(sin(xyangle), cos(zangle)))
		item.momz = FixedMul(momentum, (sin(zangle) * 2) / 3)
		
		item.sprite = states[mobjinfo[type].spawnstate].sprite
		item.frame = states[mobjinfo[type].spawnstate].frame
		if source.type == MT_PLAYER then item.scale = source.scale end
		
		item.renderflags = RF_PAPERSPRITE
		item.spritexscale = spritescale
		item.spriteyscale = spritescale
		
		item.angle = source.angle + ANGLE_90
		item.friction = FRACUNIT/4
		
		if antigravity then
			item.flags2 = $ | MF2_OBJECTFLIP
			if source.type != MT_PLAYER then item.momz = -$ end
		end
		
		item.realtype = type
		item.tosstimer = source.type == MT_PLAYER and 56 or 14
	end,
	
	-- Backup contents of hotbar
	backuphotbar = function(player)
		local hotbar = player.steve.persist.hotbar
		if mc_useluabanks then
			for slot = 0, 8 do
				local item, amount = hotbar[slot].item, hotbar[slot].amount
				if item != MT_NULL and item > 0 and item < 2048 and amount > 0 and amount <= 64 then
					luabanks[slot] = (item << 6) | (amount - 1)
				else
					luabanks[slot] = 0
				end
			end
		else
			local hotbarbackup = player.steve.persist.hotbarbackup
			for slot = 0, 8 do
				hotbarbackup[slot].item = hotbar[slot].item
				hotbarbackup[slot].amount = hotbar[slot].amount
			end
		end
	end,
	
	-- Restore contents of hotbar from backup
	restorehotbar = function(player)
		local hotbar = player.steve.persist.hotbar
		if mc_useluabanks then
			for slot = 0, 8 do
				if luabanks[slot] != 0 then
					hotbar[slot].item = luabanks[slot] >> 6
					hotbar[slot].amount = (luabanks[slot] & ~(2047 << 6)) + 1
				else
					hotbar[slot] = {item = MT_NULL, amount = 0}
				end
			end
		else
			local hotbarbackup = player.steve.persist.hotbarbackup
			for slot = 0, 8 do
				hotbar[slot].item = hotbarbackup[slot].item
				hotbar[slot].amount = hotbarbackup[slot].amount
			end
		end
	end,
	
	// Remove all items from hotbar
	clearhotbar = function(player)
		for slot = 0, 8 do
			player.steve.persist.hotbar[slot].amount = 0
		end
	end,
	
	/*
		Below is a glossary of variables for this stupid dumb animation function that is
		tailored specifically for back-and-forth animations, of which there are two of them
		
		progress		position of the animation in angle form
		progressadd		progress is incremented by this, multiplied by modifier
		extent			current allowed variation of the animation from its starting point;
						unless immediate is true, value is updated only when progress reaches
						a multiple of 180 degrees (equivalent to starting point)
		realextent		the animation's "real" variation, used to update extent when needed
		extendadd		realextent is incremented by this, or decremented if moving is false
		minextent		minimum allowed variation of the animation;
						used when the animation has just started, but not when it is ending
		maxextent		maximum allowed variation of the animation;
						multiplied by modifier only if modifier is below FRACUNIT, to allow
						fine-grained control of the variation while keeping it in-bounds
		
		animation		the animation being updated
		modifier		progressadd (and maxextent if below FRACUNIT) is multiplied by this
		moving			if true, animation progresses normally;
						if false, extent is steadily decreased to zero, stopping the animation
		immediate		if true, extent always has the same value as realextent
		
		The current variation will be decremented by double the value of extentadd if/while it
		exceeds the maximum variation, in order to quickly correct it to the new maximum
	*/
	animate = function(animation, modifier, moving, immediate)
		local progressadd = FixedMul(animation.progressadd, modifier)
		local maxextent = animation.maxextent
		local extentadd = animation.extentadd
		
		if modifier < FRACUNIT then
			maxextent = min($, FixedMul($, modifier))
		end
		
		if animation.extent > maxextent then
			moving = false
			extentadd = $ * 2
		end
		
		if moving then
			animation.realextent = min(maxextent, max(animation.minextent, $) + extentadd)
			
			local oldsign = animation.progress != 0 and animation.progress / abs(animation.progress) or 0
			animation.progress = $ + progressadd
			local newsign = animation.progress != 0 and animation.progress / abs(animation.progress) or 0
			
			if oldsign != newsign or immediate then animation.extent = animation.realextent end
		elseif animation.extent > 0 then
			animation.realextent = max(0, $ - extentadd)
			animation.progress = $ + progressadd
			animation.extent = animation.realextent
		else
			animation.realextent = 0
			animation.progress = 0
			animation.extent = 0
		end
	end
})