addHook("PreThinkFrame", do
	for player in players.iterate do
		if not stevehelper.validinfo(player) then continue end
		local steve = player.steve
		local last = steve.play.last
		
		-- Decrement timers each frame
		if steve.play.hurttimer > 0 then steve.play.hurttimer = $ - 1 end
		if steve.play.redtimer > 0 then steve.play.redtimer = $ - 1 end
		if steve.play.blinktimer > 0 then steve.play.blinktimer = $ - 1 end
		if steve.play.soundtimer > 0 then steve.play.soundtimer = $ - 1 end
		if steve.play.regentimer > 0 then steve.play.regentimer = $ - 1 end
		if steve.play.usetimer > 0 then steve.play.usetimer = $ - 1 end
		
		-- Stop all movement if dead, and drop items if keepinventory is not enabled
		if steve.play.dying then
			if CV_FindVar("mc_keepinventory").value == 0 then
				local hotbar = steve.persist.hotbar
				for slot = 0, 8 do
					if hotbar[slot].item != MT_NULL and hotbar[slot].amount > 0 then
						for i = 1, hotbar[slot].amount do
							stevehelper.dropitem(player.mo, hotbar[slot].item, FixedAngle(((360*FRACUNIT) / 9) * slot), ANGLE_45, 8*FRACUNIT, steve.play.antigravity)
						end
						hotbar[slot].amount = 0
					end
				end
			end
			continue
		end
		
		-- Decrement oxygen level while underwater
		if player.mo.state == S_PLAY_GASP then
			steve.play.breath = 20
			steve.play.breathoffset = leveltime % 35
		elseif steve.play.submerged and player.exiting == 0 and not (player.pflags & PF_GODMODE) then
			player.powers[pw_underwater] = 1050
			if leveltime % 35 == steve.play.breathoffset then
				if steve.play.breath >= 0 then
					steve.play.breath = $ - 1
					steve.play.poptimer = 5
				else
					stevehelper.damage(player, nil, 2)
				end
			end
		end
		if steve.play.poptimer > 0 then steve.play.poptimer = $ - 1 end
		
		-- Change hotbar slot, and adapt to scroll wheel controls if necessary
		if consoleplayer != nil and #consoleplayer == #player and (player.cmd.buttons & (BT_WEAPONNEXT | BT_WEAPONPREV)) then
			local keyname, amount, hold = "", 0, 0
			if player.cmd.buttons & BT_WEAPONNEXT then
				keyname = input.keyNumToName(input.gameControlToKeyNum(GC_WEAPONNEXT))
				amount = 1
				hold = player.lastbuttons & BT_WEAPONNEXT
			elseif player.cmd.buttons & BT_WEAPONPREV then
				keyname = input.keyNumToName(input.gameControlToKeyNum(GC_WEAPONPREV))
				amount = -1
				hold = player.lastbuttons & BT_WEAPONPREV
			end
			if keyname == "wheel 1 up" or keyname == "wheel 1 down" or not hold then
				COM_BufInsertText(player, "mc_shiftslot " .. amount)
			end
		end
		
		-- Perform Minecraft-style jump when jump button is pressed
		if (P_IsObjectOnGround(player.mo) or (player.powers[pw_carry] == CR_ROPEHANG and not (player.lastbuttons & BT_JUMP))) and (player.cmd.buttons & BT_JUMP) then
			local jumpforce = 15*FRACUNIT
			if player.mo.eflags & MFE_UNDERWATER then
				jumpforce = $ / 3
			end
			if player.powers[pw_carry] == CR_ROPEHANG then
				jumpforce = ($ * 5) / 4
				player.powers[pw_carry] = CR_NONE
			end
			P_SetObjectMomZ(player.mo, jumpforce)
			player.pflags = $ | PF_JUMPDOWN | PF_JUMPED | PF_NOJUMPDAMAGE
		elseif not P_IsObjectOnGround(player.mo) then
			player.pflags = $ & ~PF_JUMPDOWN
		else
			player.pflags = $ & ~(PF_JUMPDOWN | PF_JUMPED | PF_NOJUMPDAMAGE)
		end
		
		-- Update player states
		steve.play.moving = player.cmd.forwardmove != 0 or player.cmd.sidemove != 0
		steve.play.sneaking = (player.cmd.buttons & BT_CUSTOM1) and not (steve.play.sitting or steve.play.hanging)
		if steve.play.sprinting or (player.cmd.buttons & BT_SPIN) then
			steve.play.sprinting = 
				not steve.play.sneaking and player.cmd.forwardmove > 0
				and FixedDiv(player.speed, player.mo.scale) > FRACUNIT and not player.powers[pw_pushing]
		end
		steve.play.bhopping = (player.pflags & PF_JUMPED) and steve.play.sprinting
		steve.play.sitting = player.powers[pw_carry] == CR_MACESPIN or player.powers[pw_carry] == CR_MINECART
		steve.play.hanging = player.powers[pw_carry] == CR_PLAYER or player.powers[pw_carry] == CR_ROPEHANG
		steve.play.tossing = (player.cmd.buttons & BT_CUSTOM2) and not (player.lastbuttons & BT_CUSTOM2)
		steve.play.underwater = (player.mo.eflags & MFE_UNDERWATER) and not (player.mo.eflags & MFE_GOOWATER)
		steve.play.goopwater = player.mo.eflags & MFE_GOOWATER
		steve.play.submerged = (player.mo.eflags & MFE_UNDERWATER) and not (player.mo.eflags & MFE_TOUCHWATER)
		steve.play.antigravity = player.mo.eflags & MFE_VERTICALFLIP
		
		-- Toss selected item if toss button is pressed
		if steve.play.tossing and player.exiting == 0 then
			local slot = steve.persist.hotbar[steve.persist.hotbarslot]
			if slot.item != MT_NULL and slot.amount > 0 then
				stevehelper.dropitem(player.mo, slot.item, player.mo.angle, player.aiming, 20*FRACUNIT, steve.play.antigravity)
				if CV_FindVar("mc_itemfreeze").value == 0 then slot.amount = $ - 1 end
			end
		end
		
		-- Cushion Z momentum upon entering a liquid
		if not (last.underwater or last.goopwater) and (steve.play.underwater or steve.play.goopwater) then
			player.mo.momz = ($ * 4) / 7
		end
		
		-- Begin depleting oxygen if fully submerged in a liquid
		if not last.submerged and steve.play.submerged then
			steve.play.breath = 20
			steve.play.breathoffset = leveltime % 35
		end
		
		if steve.play.underwater then
			-- Swim upwards if jump button is held while underwater
			local realmomz = FixedDiv(player.mo.momz * (steve.play.antigravity and -1 or 1), player.mo.scale)
			if (player.cmd.buttons & BT_JUMP) then
				if realmomz > 0 then
					P_SetObjectMomZ(player.mo, min(5*FRACUNIT, realmomz + FRACUNIT/3))
				elseif realmomz <= 0 then
					P_SetObjectMomZ(player.mo, (FRACUNIT*2)/3, true)
				end
			elseif realmomz > 0 then
				P_SetObjectMomZ(player.mo, -FRACUNIT/3, true)
			end
		elseif last.underwater and (player.cmd.buttons & BT_JUMP) and player.powers[pw_pushing] then
			-- Jump out of water if swimming to surface while moving against a wall
			P_SetObjectMomZ(player.mo, 15*FRACUNIT)
		end
		
		-- Increase gravitational pull
		if not P_IsObjectOnGround(player.mo) and not (player.mo.eflags & MFE_UNDERWATER) and not player.powers[pw_carry] then
			player.mo.momz = $ + P_GetMobjGravity(player.mo) * 2
		end
		
		-- Increment fist swinging animation position
		if steve.play.usepos > 0 and steve.play.usepos < steve.play.usemax then
			steve.play.usepos = ($ + 1) % steve.play.usemax
		end
		
		-- Initiate mining/hitting/placing actions only when Steve's fist is down
		if steve.play.usetimer == 0 then
			if not steve.play.punchactive and (player.cmd.buttons & BT_ATTACK) then
				steve.play.usepos = 1
				steve.play.usetimer = 7
				steve.play.punchaction = true
				steve.play.punchactive = true
			elseif steve.play.punchactive and not (player.cmd.buttons & BT_ATTACK) then
				steve.play.punchactive = false
			end
			if not steve.play.placeactive and (player.cmd.buttons & BT_FIRENORMAL) then
				steve.play.placeaction = true
				steve.play.placeactive = true
			elseif steve.play.placeactive and not (player.cmd.buttons & BT_FIRENORMAL) then
				steve.play.placeactive = false
			end
		end
		
		-- Increase force of springs, breaking minecarts, and dust devils
		if (player.mo.eflags & MFE_SPRUNG)
		or last.minecart and player.powers[pw_carry] != CR_MINECART
		or last.dustdevil and player.powers[pw_carry] != CR_DUSTDEVIL then
			player.mo.momx = ($ * 7) / 4
			player.mo.momy = ($ * 7) / 4
			player.mo.momz = ($ * 29) / 16
		end
		
		-- Increase force of mace spins
		if last.macespin and player.powers[pw_carry] != CR_MACESPIN then
			player.mo.momx = ($ * 7) / 2
			player.mo.momy = ($ * 7) / 2
			player.mo.momz = ($ * 29) / 16
		end
		
		-- Backup relevant states/properties for use in subsequent tic
		last.x = player.mo.x
		last.y = player.mo.y
		last.underwater = steve.play.underwater
		last.goopwater = steve.play.goopwater
		last.submerged = steve.play.submerged
		last.minecart = player.powers[pw_carry] == CR_MINECART
		last.dustdevil = player.powers[pw_carry] == CR_DUSTDEVIL
		last.macespin = player.powers[pw_carry] == CR_MACESPIN
	end
end)

addHook("ThinkFrame", do
	for player in players.iterate do
		if not stevehelper.validinfo(player) or player.steve.play.dying then continue end
		local steve = player.steve
		local last = steve.play.last
		
		if stevehelper.validmodel(player) then
			-- Update position of Steve model
			steve.model.x = player.mo.x
			steve.model.y = player.mo.y
			steve.model.z = player.mo.z
			
			-- Update angle of Steve's head
			steve.model.xyangle = player.mo.angle
			steve.model.zangle = player.aiming
			
			-- Recreate Minecraft's body turning behavior
			local turned = false
			if abs(player.rmomx) > FRACUNIT/4 or abs(player.rmomy) > FRACUNIT/4 then
				local direction = max(-FRACUNIT, min(FRACUNIT, (sin(R_PointToAngle2(0, 0, player.rmomx, player.rmomy) - steve.model.xyangle) * 3) / 2))
				local destination = steve.model.xyangle + FixedAngle(direction * 45)
				local distance = destination - steve.model.bodyxyangle
				if steve.model.bodyxyangle + distance == destination then
					steve.model.bodyxyangle = $ + distance / 3
					turned = true
				end
			end
			if not turned then
				local distance = steve.model.xyangle - steve.model.bodyxyangle
				if abs(distance) > ANGLE_45 then
					steve.model.bodyxyangle = $ + (abs(distance) - ANGLE_45) * (distance / abs(distance))
				end
			end
			
			-- Disable first-person camera bob and adjust while sneaking
			local viewz = FixedMul(steve.model.scale, player.mo.scale) * (steve.play.sneaking and 102 or 112)
			player.viewz = player.mo.z + (steve.play.antigravity and player.mo.height or 0) + viewz * (steve.play.antigravity and -1 or 1)
		end
		
		-- Update maximum speed according to current state
		player.normalspeed = stevehelper.maxspeed(player)
		player.acceleration = stevehelper.maxspeed(player)/FRACUNIT
		
		-- Heal Steve and back up hotbar contents upon touching a star post
		if player.starposttime == leveltime and abs(steve.persist.spawntime - leveltime) > 3 then
			local starpostheal = CV_FindVar("mc_starpostheal").value
			steve.persist.health = min(20, $ + starpostheal)
			if starpostheal > 0 then steve.play.regentimer = 20 end
			if not netgame and CV_FindVar("mc_dobackups").value == 1 then
				stevehelper.backuphotbar(player)
			end
		end
		
		-- Construct animation speed modifier
		local modifier = FRACUNIT
		if steve.play.sneaking   then modifier = $ / 4 end
		if steve.play.sprinting  then modifier = ($ * 7) / 6 end
		if steve.play.underwater then modifier = ($ * 2) / 3 end
		
		-- Update progress of bobbing and arm swinging animations
		local airborne = not P_IsObjectOnGround(player.mo) and not (player.pflags & PF_JUMPED)
		if player.speed > 0 and (steve.play.moving and not player.powers[pw_carry]) or airborne then
			if not player.powers[pw_pushing] and P_IsObjectOnGround(player.mo) then
				stevehelper.animate(steve.play.bob, modifier, true, true)
			elseif player.powers[pw_pushing] or airborne then
				modifier = min(FRACUNIT, FixedMul($, FixedDiv(stevehelper.currentspeed(player), player.normalspeed)))
			end
			stevehelper.animate(steve.play.swing, modifier, true, false)
		else
			stevehelper.animate(steve.play.swing, modifier, false, false)
			stevehelper.animate(steve.play.bob, modifier, false, true)
		end
		
		-- Increase FOV while sprinting
		if steve.play.sprinting and not (steve.play.sitting or steve.play.hanging) then
			local sprintfov = CV_FindVar("mc_sprintfov").value * FRACUNIT
			steve.play.currentfov = min(sprintfov, $ + (sprintfov - $) / 3)
			player.fovadd = steve.play.currentfov
		elseif steve.play.currentfov > 0 then
			steve.play.currentfov = ($ * 3) / 4
			player.fovadd = steve.play.currentfov
		end
		
		-- Prevent player from moving off of cliff edges while sneaking
		if steve.play.sneaking and P_IsObjectOnGround(player.mo) and last.x != nil and last.y != nil then
			if not steve.play.antigravity then
				if player.mo.floorz - P_FloorzAtPos(player.mo.x, last.y, player.mo.z, player.mo.height) > MAXSTEPMOVE then
					P_MoveOrigin(player.mo, last.x, player.mo.y, player.mo.z)
				end
				if player.mo.floorz - P_FloorzAtPos(last.x, player.mo.y, player.mo.z, player.mo.height) > MAXSTEPMOVE then
					P_MoveOrigin(player.mo, player.mo.x, last.y, player.mo.z)
				end
			else
				if player.mo.ceilingz - P_CeilingzAtPos(player.mo.x, last.y, player.mo.z, player.mo.height) < -MAXSTEPMOVE then
					P_MoveOrigin(player.mo, last.x, player.mo.y, player.mo.z)
				end
				if player.mo.ceilingz - P_CeilingzAtPos(last.x, player.mo.y, player.mo.z, player.mo.height) < -MAXSTEPMOVE then
					P_MoveOrigin(player.mo, player.mo.x, last.y, player.mo.z)
				end
			end
		end
		
		-- Only cast rays for other players when they are performing an action
		-- Additionally, don't cast rays for any players that have completed the level
		if player.exiting != 0 or (consoleplayer != nil and #consoleplayer != #player and not (steve.play.punchactive or steve.play.placeactive)) then
			continue
		end
		
		local rayx, rayy, rayz = player.mo.x, player.mo.y, player.viewz - (18*FRACUNIT)
		local raymomx = FixedMul(player.mo.scale, FixedMul(cos(player.mo.angle), cos(player.aiming)))
		local raymomy = FixedMul(player.mo.scale, FixedMul(sin(player.mo.angle), cos(player.aiming)))
		local raymomz = FixedMul(player.mo.scale, sin(player.aiming))
		
		local ray = P_SpawnMobj(rayx, rayy, rayz, MT_MCRAY)
		ray.momx, ray.momy, ray.momz = raymomx, raymomy, raymomz
		ray.drawonlyforplayer = player
		
		local maxdist = 256
		
		-- Cast ray to detect objects that can be hit/mined
		for dist = 1, maxdist do
			P_SceneryXYMovement(ray)
			P_SceneryZMovement(ray)
			
			if not ray.valid then return end
			
			if ray.collideobj != nil then
				if ray.collideobj.flags & (MF_ENEMY | MF_BOSS) then
					ray.canhit = true
					maxdist = dist
					break
				elseif ray.collideobj.flags & (MF_SOLID | MF_SPRING) then
					ray.canmine = true
					maxdist = dist
					break
				end
			elseif ray.z == ray.floorz then
				for fof in ray.subsector.sector.ffloors() do
					if (fof.flags & FF_EXISTS) and (fof.flags & FF_BUSTUP) and ray.z == fof.topheight then
						ray.bustfof = fof
					end
				end
			end
		end
		
		local slot = steve.persist.hotbar[steve.persist.hotbarslot]
		local empty = slot.item == MT_NULL or slot.amount <= 0
		
		if not empty then
			P_SetOrigin(ray, rayx, rayy, rayz)
			ray.momx, ray.momy, ray.momz = raymomx, raymomy, raymomz
			ray.radius = mobjinfo[slot.item].radius
			ray.height = mobjinfo[slot.item].height
			
			-- Cast ray to detect surfaces that can be placed on
			for dist = 1, maxdist do
				P_SceneryXYMovement(ray)
				P_SceneryZMovement(ray)
				
				if ray.momx != raymomx or ray.momy != raymomy or ray.momz != raymomz then
					ray.canplace = not (
						(mobjinfo[slot.item].flags & MF_SOLID) and R_PointToDist2(rayx, rayy, ray.x, ray.y) <= player.mo.radius * 2
						and ray.z > player.mo.z - ray.height and ray.z < player.mo.z + player.mo.height
					)
					break
				end
			end
		end
		
		local permitplace = not empty and (ray.canmine or ray.canplace)
		
		-- Show "ghost" of object as it would appear if it were placed
		if permitplace then
			ray.sprite = states[mobjinfo[slot.item].spawnstate].sprite
			ray.frame = states[mobjinfo[slot.item].spawnstate].frame | FF_TRANS50
			if steve.play.antigravity then ray.eflags = MFE_VERTICALFLIP end
			ray.angle = player.mo.angle
		end
		
		-- Always allow player to mine solid objects by holding the punch button
		if steve.play.punchactive and ray.canmine and not steve.play.punchaction then
			steve.play.punchactive = false
		end
		
		-- Always allow player to place objects by holding the place button
		if steve.play.placeactive and ray.canplace and not steve.play.placeaction then 
			steve.play.placeactive = false
		end
		
		-- Damage object if hittable, or mine object if mineable
		if steve.play.punchaction then
			if ray.canhit then
				P_DamageMobj(ray.collideobj, player.mo, player.mo)
			elseif ray.canmine then
				if ray.collideobj.breakage == nil then
					ray.collideobj.breakage = FixedSqrt(FixedMul(ray.collideobj.radius * 2, ray.collideobj.height)) / 5
				else
					ray.collideobj.breakage = $ - FRACUNIT
				end
				if ray.collideobj.breakage <= 0 or CV_FindVar("mc_instamine").value == 1 then
					stevehelper.dropitem(ray.collideobj, ray.collideobj.type, 0, ANGLE_90, 8*FRACUNIT, steve.play.antigravity)
					P_RemoveMobj(ray.collideobj)
					S_StartSound(player.mo, sfx_mcbrk1 + P_RandomKey(4))
				else
					S_StartSound(player.mo, sfx_mcmne1 + P_RandomKey(2))
				end
			elseif ray.bustfof then
				EV_CrumbleChain(ray.bustfof)
			end
			steve.play.punchaction = false
		end
		
		-- Place held object
		if steve.play.placeaction then
			if permitplace then
				local spawnedobj = P_SpawnMobj(ray.x, ray.y, ray.z, slot.item)
				spawnedobj.angle = player.mo.angle
				spawnedobj.flags = $ | MF_NOGRAVITY
				if steve.play.antigravity then spawnedobj.eflags = $ | MFE_VERTICALFLIP end
				
				if CV_FindVar("mc_itemfreeze").value == 0 then slot.amount = $ - 1 end
				steve.play.usepos = 1
				steve.play.usetimer = 5
				
				S_StartSound(player.mo, sfx_mcbrk1 + P_RandomKey(4))
			end
			steve.play.placeaction = false
		end
	end
end)

addHook("PostThinkFrame", do
	for player in players.iterate do
		if player.mo == nil then continue end
		if player.mo.skin == "mcsteve" then
			-- If Steve has just died in a netgame, destroy model/health data so it can be recreated
			-- However, we need to wait a frame to recreate the data or else the dontdrawforviewmobj property breaks
			if netgame and player.playerstate == PST_REBORN then
				stevehelper.destroy(player, 2)
				continue
			end
			-- If Steve's model is invalid, destroy it so it can be recreated
			if not stevehelper.validmodel(player) then
				stevehelper.destroy(player, 0)
			end
			-- Initialize any missing/destroyed Steve attributes
			if player.steve == nil or player.steve.model == nil or player.steve.play == nil or player.steve.persist == nil then
				local restore = not netgame and player.steve == nil
				stevehelper.create(player, (FRACUNIT*7)/8)
				if restore then stevehelper.restorehotbar(player) end
				if consoleplayer != nil and #consoleplayer == #player then
					hud.disable("score")
					hud.disable("time")
					hud.disable("rings")
					hud.disable("lives")
				end
			end
			-- Back up hotbar contents upon completing a level in singleplayer
			if not netgame and player.exiting == 2 and CV_FindVar("mc_dobackups") then
				stevehelper.backuphotbar(player)
			end
			-- Update Steve model each frame
			player.mo.sprite2 = SPR2_NULL
			stevehelper.update(player)
		elseif skins[player.skin].name != "mcsteve" and player.steve != nil then
			-- Restore player to default state if skin is changed away from Steve
			stevehelper.destroy(player, 4)
			-- Disable spawnkilling protection if active
			player.mo.flags = $ & ~MF_NOCLIPTHING
		end
	end
end)

addHook("MapChange", function(mapnum)
	for player in players.iterate do
		if player.steve == nil or player.steve.persist == nil then continue end
		if not netgame then
			if gamemap == mapnum then
				if player.deadtimer == 0 then
					-- Prevent retries from decrementing lives counter
					player.lives = $ + 1
				else
					-- Destroy model/health data upon death
					-- This resolves an edge case where retrying during the death animation causes Steve to enter a bugged state 
					stevehelper.destroy(player, 2)
				end
				-- Restore hotbar contents with map reset
				if CV_FindVar("mc_keepinventory").value == 1 then
					stevehelper.restorehotbar(player)
				end
			end
		end
		-- Make sure player properties do or don't persist upon changing the map
		if netgame or player.deadtimer == 0 then
			if CV_FindVar("mc_healthpersist").value == 0 then
				player.steve.persist.health = 20
			end
			if CV_FindVar("mc_itempersist").value == 0 then
				stevehelper.clearhotbar(player)
			end
		end
	end
end)

addHook("PlayerThink", function(player)
	if player.mo == nil or skins[player.skin].name != "mcsteve" then return end
	
	-- If Steve has just respawned in multiplayer, make him temporarily intangible to objects to prevent spawnkilling
	if netgame then
		if player.powers[pw_flashing] then
			player.mo.flags = $ | MF_NOCLIPTHING
		elseif not player.powers[pw_flashing] then
			player.mo.flags = $ & ~MF_NOCLIPTHING
		end
	end
	
	local map = mapheaderinfo[gamemap]
	
	-- Work around issue where Steve's height causes him to be forced into the shuttle in BCZ3
	if (leveltime == 0 or player.playerstate == PST_REBORN) and map.lvlttl == "Black Core" and map.actnum == 3 then
		player.mo.z = 1920*FRACUNIT
	end
	
	-- Change Steve's height in NiGHTS stages to fix collision and camera angle
	if player.nightstime != 0 then
		player.mo.height = skins[player.mo.skin].height
		player.height = player.mo.height
	elseif player.height != skins[player.skin].height then
		player.height = skins[player.skin].height
	end
end)

-- Get spawn time for use in comparisons
addHook("PlayerSpawn", function(player)
	if player.steve != nil and player.steve.persist != nil then
		player.steve.persist.spawntime = leveltime
	end
end)

-- Destroy Steve model and properties upon quitting netgame
addHook("PlayerQuit", function(player)
	if player.steve != nil then stevehelper.destroy(player, 4) end
end)

-- Synchronize each player's Steve properties
addHook("NetVars", function(network)
	for player in players.iterate do player.steve = network($) end
end)

-- Initiate Minecraft-style damage behavior upon taking damage
addHook("MobjDamage", function(target, inflictor)
	if stevehelper.validinfo(target.player) then
		if target.player.steve.play.hurttimer == 0 then
			if inflictor != nil and inflictor.valid then
				-- Damage player by an amount based on the inflictor's size
				stevehelper.damage(target.player, inflictor, max(1, FixedSqrt(FixedMul(inflictor.radius * 2, inflictor.height) / 200)/FRACUNIT))
			else
				-- If the inflictor isn't an object (ie. lava), deal 1 1/2 hearts of damage
				stevehelper.damage(target.player, nil, 3)
			end
		end
		return true
	end
end, MT_PLAYER)

-- Initiate Minecraft-style death animation upon dying
addHook("MobjDeath", function(target)
	if stevehelper.validinfo(target.player) then
		local player = target.player
		
		player.steve.play.dying = true
		player.mo.momx, player.mo.momy, player.mo.momz = 0, 0, 0
		
		S_StartSound(player.mo, sfx_mcoof1 + P_RandomKey(3))
		
		return true
	end
end, MT_PLAYER)

-- Check if ray has collided with an object
addHook("MobjMoveCollide", function(ray, object)
	if ray.collideobj == nil and object.info.doomednum != -1 and ray.z >= object.z and ray.z < object.z + object.height then
		ray.collideobj = object
	end
end, MT_MCRAY)

-- Check if ray has collided with a bustable block
addHook("MobjLineCollide", function(ray, line)
	for _, sector in ipairs({line.frontsector, line.backsector}) do
		for fof in sector.ffloors() do
			if (fof.flags & FF_EXISTS) and (fof.flags & FF_BUSTUP) and ray.z + ray.height > fof.bottomheight and ray.z < fof.topheight then
				ray.bustfof = fof
			end
		end
	end
end, MT_MCRAY)

-- Force the Metal Sonic boss to spin out upon collision
-- Because otherwise it just... doesn't happen? For some reason???
addHook("MobjCollide", function(playerobj, object)
	if stevehelper.validinfo(playerobj.player) and not playerobj.player.steve.play.dying
	and object.type == MT_METALSONIC_BATTLE and object.state == S_METALSONIC_FLOAT
	and object.z > playerobj.z - object.height and object.z < playerobj.z + playerobj.height then
		object.flags2 = MF2_CLASSICPUSH
	end
end, MT_PLAYER)

addHook("MobjThinker", function(item)
	-- Move at double speed along the Z axis to simulate Minecraft's gravity
	P_ZMovement(item)
	if not item.valid then return end

	-- Check for nearby players, and move item to hotbar if close enough
	if item.tosstimer == 0 then
		local pickup = false
		local maxxydist = item.radius * 5
		local maxzdist = item.height * 3
		
		searchBlockmap("objects", function(item, object)
			if item.attractplayer == nil and object.valid and object.type == MT_PLAYER
			and stevehelper.validinfo(object.player) and not object.player.steve.play.dying then
				local playerzadjust = object.z + object.height / 2
				local itemzadjust = item.z + (object.player.steve.play.antigravity and item.height or 0)
				local xydist = R_PointToDist2(item.x, item.y, object.x, object.y)
				local zdist = abs(itemzadjust - playerzadjust)
				if xydist <= maxxydist and zdist <= maxzdist then
					item.attractplayer = object.player
				end
			end
		end, item, item.x - maxxydist, item.x + maxxydist, item.y - maxxydist, item.y + maxxydist)
		
		if item.attractplayer != nil then
			local player = item.attractplayer
			local antigravity = player.steve.play.antigravity
			local playerzadjust = player.mo.z + player.height / 2
			local itemzadjust = item.z + (antigravity and item.height or 0)
			local attractstart, clippoint = 4, 1
			if item.attracttimer == nil then
				if not stevehelper.giveitem(player, item.realtype) then return end
				item.initx, item.inity, item.initz = item.x, item.y, itemzadjust
				item.attracttimer = attractstart
				item.segments = attractstart + clippoint
				item.flags = $ | MF_NOCLIP | MF_NOCLIPHEIGHT | MF_NOGRAVITY
			elseif item.attracttimer > 0 then
				item.attracttimer = $ - 1
				item.spritexscale = $ - ($ / item.segments)
				item.spriteyscale = $ - ($ / item.segments)
			else
				P_RemoveMobj(item)
				return
			end
			item.position = item.attracttimer + clippoint
			P_MoveOrigin(item,
				player.mo.x + ((item.initx - player.mo.x) / item.segments) * item.position,
				player.mo.y + ((item.inity - player.mo.y) / item.segments) * item.position,
				playerzadjust + ((item.initz - playerzadjust) / item.segments) * (item.position - (antigravity and item.segments or 0))
			)
		end
	end
	
	-- Decrement pickup timer each frame
	item.tosstimer = max(0, $ - 1)
	-- Continuously rotate item
	item.angle = $ + ANG1
	
	-- Bob up and down while on the ground
	if P_IsObjectOnGround(item) then
		if item.bobangle == nil then
			item.bobangle = 0
		else
			item.bobangle = $ + ANG1
		end
		item.spriteyoffset = FixedDiv((-cos(item.bobangle * 3) + FRACUNIT) * 10, item.spriteyscale)
	end
end, MT_MCITEM)

-- Re-enable HUD if quitting out of gameplay
addHook("GameQuit", function(quitting)
	if not quitting and skins[consoleplayer.skin].name == "mcsteve" then
		hud.enable("score")
		hud.enable("time")
		hud.enable("rings")
		hud.enable("lives")
	end
end)